#include <cpse/math/matrix.tpp>

namespace cpse {
namespace math {

typedef basic_matrix<3, 3> m3;
typedef basic_matrix<3, 1> v3;

void matrix3_assign_handler::go(m3 &d, matrix_binop<binop_mul, m3, m3> s)
{
	for (int i=0; i!=3; ++i) {

		d(i,0) = s.a(i,0)*s.b(0,0) + s.a(i,1)*s.b(1,0)
			+ s.a(i,2)*s.b(2,0);

		d(i,1) = s.a(i,0)*s.b(0,1) + s.a(i,1)*s.b(1,1)
			+ s.a(i,2)*s.b(2,1);

		d(i,2) = s.a(i,0)*s.b(0,2) + s.a(i,1)*s.b(1,2)
			+ s.a(i,2)*s.b(2,2);
	}
}

float det_bm(const m3 &m)
{
	return m(0,0) * det(cofactor<0,0>(m))
		- m(1,0) * det(cofactor<1,0>(m))
		+ m(2,0) * det(cofactor<2,0>(m));
}

void matrix3_assign_handler::go(m3 &dest, matrix_unop<unop_adjoint, m3> src)
{
	dest(0,0) = +det(cofactor<0,0>(src.a));
	dest(1,0) = -det(cofactor<0,1>(src.a));
	dest(2,0) = +det(cofactor<0,2>(src.a));

	dest(0,1) = -det(cofactor<1,0>(src.a));
	dest(1,1) = +det(cofactor<1,1>(src.a));
	dest(2,1) = -det(cofactor<1,2>(src.a));

	dest(0,2) = +det(cofactor<2,0>(src.a));
	dest(1,2) = -det(cofactor<2,1>(src.a));
	dest(2,2) = +det(cofactor<2,2>(src.a));
}

void matrix3_assign_handler::go(m3 &dest, matrix_unop<unop_inverse, m3> src)
{
	float det_src = det_bm(src.a);
//	if (close(det_src, 0.f))
//		throw std::logic_error("inverse of singular matrix requested");

	float d = 1.f / det_src;

	// dest = adjoint(src.a) / det(src.a)
	dest(0,0) = d * +det(cofactor<0,0>(src.a));
	dest(1,0) = d * -det(cofactor<0,1>(src.a));
	dest(2,0) = d * +det(cofactor<0,2>(src.a));

	dest(0,1) = d * -det(cofactor<1,0>(src.a));
	dest(1,1) = d * +det(cofactor<1,1>(src.a));
	dest(2,1) = d * -det(cofactor<1,2>(src.a));

	dest(0,2) = d * +det(cofactor<2,0>(src.a));
	dest(1,2) = d * -det(cofactor<2,1>(src.a));
	dest(2,2) = d * +det(cofactor<2,2>(src.a));
}

} // namespace math
} // namespace cpse
