
namespace cpse {
namespace math {

// identity
inline void assign(basic_matrix<3, 3> &m, identity_t)
{
	m(0,1) = m(0,2) = 0.f;
	m(1,0) = m(1,2) = 0.f;
	m(2,0) = m(2,1) = 0.f;
	m(0,0) = m(1,1) = m(2,2) = 1.f;
}

// zero
inline void assign(basic_matrix<3, 1> &d, zero_t)
	{ d.x = d.y = d.z = 0.f; }

struct matrix3_assign_handler : matrix_assign_handler {
	using matrix_assign_handler::go;
	typedef basic_matrix<3, 3> m3;
	typedef basic_matrix<3, 1> v3;

	// vector (cross) product
	static void go(v3 &dest, matrix_binop<binop_mul, v3, v3> src)
	{
		dest.x = src.a.y*src.b.z - src.a.z*src.b.y;
		dest.y = src.a.z*src.b.x - src.a.x*src.b.z;
		dest.z = src.a.x*src.b.y - src.a.y*src.b.x;
	}

	// explicit vector (cross) product
	static void go(v3 &dest, matrix_binop<binop_cross, v3, v3> src)
	{
		dest.x = src.a.y*src.b.z - src.a.z*src.b.y;
		dest.y = src.a.z*src.b.x - src.a.x*src.b.z;
		dest.z = src.a.x*src.b.y - src.a.y*src.b.x;
	}

	// inverse
	static void go(m3 &, matrix_unop<unop_inverse, m3>);

	// adjoint
	static void go(m3 &, matrix_unop<unop_adjoint, m3>);

	// normalize vector
	inline static void go(v3 &dest, matrix_unop<unop_normalize, v3> src) {
		float l = sqrt(length_squared_bm(src.a));
		dest.x = src.a.x / l;
		dest.y = src.a.y / l;
		dest.z = src.a.z / l;
	}

	// expand vector
	static void go(v3 &dest, matrix_binop<binop_expand,
		basic_matrix<2, 1>, float_keeper> src)
	{
		dest.x = src.a.x;
		dest.y = src.a.y;
		dest.z = src.b.f;
	}

	// perspective division
	static void go(v3 &dest, matrix_unop<unop_wdiv,
		basic_matrix<4, 1> > src)
	{
		dest.x = src.a.x / src.a.w;
		dest.y = src.a.y / src.a.w;
		dest.z = src.a.z / src.a.w;
	}

	// ordinary matrix multiplication
	static void go(m3 &, matrix_binop<binop_mul, m3, m3>);
	inline static void go(v3 &d, matrix_binop<binop_mul, m3, v3> s)
	{
		d.x = s.a(0,0)*s.b.x + s.a(0,1)*s.b.y + s.a(0,2)*s.b.z;
		d.y = s.a(1,0)*s.b.x + s.a(1,1)*s.b.y + s.a(1,2)*s.b.z;
		d.z = s.a(2,0)*s.b.x + s.a(2,1)*s.b.y + s.a(2,2)*s.b.z;
	}
};

} // namespace math
} // namespace cpse
