
namespace cpse {
namespace math {

// identity
void assign(basic_matrix<4, 4> &, identity_t);

// zero
inline void assign(basic_matrix<4, 1> &d, zero_t)
	{ d.x = 0.f; d.y = 0.f; d.z = 0.f; d.w = 0.f; }

struct matrix4_assign_handler : matrix_assign_handler {
	using matrix_assign_handler::go;
	typedef basic_matrix<4, 4> m4;
	typedef basic_matrix<4, 1> v4;

	// inverse
	static void go(m4 &, matrix_unop<unop_inverse, m4>);

	// adjoint
	static void go(m4 &, matrix_unop<unop_adjoint, m4>);

	// make translation matrix
	static void go(m4 &, matrix_unop<unop_translation,
		basic_matrix<3, 1> >);

	// make scaling matrix
	static void go(m4 &, matrix_unop<unop_scaling, basic_matrix<3, 1> >);
	static void go(m4 &, matrix_unop<unop_scaling, float_keeper>);

	// make rotation matrix
	static void go(m4 &, matrix_binop<binop_rotation, basic_matrix<3, 1>, 
		float_keeper>);
	static void go(m4 &, matrix_binop<binop_rotation, x_axis_type, 
		float_keeper>);
	static void go(m4 &, matrix_binop<binop_rotation, y_axis_type, 
		float_keeper>);
	static void go(m4 &, matrix_binop<binop_rotation, z_axis_type, 
		float_keeper>);

	// normalize vector
	inline static void go(v4 &dest, matrix_unop<unop_normalize, v4> src) {
		float l = sqrt(length_squared_bm(src.a));
		dest.x = src.a.x / l;
		dest.y = src.a.y / l;
		dest.z = src.a.z / l;
		dest.w = src.a.w / l;
	}

	// expand vector
	static void go(v4 &dest, matrix_binop<binop_expand,
		basic_matrix<3, 1>, float_keeper> src)
	{
		dest.x = src.a.x;
		dest.y = src.a.y;
		dest.z = src.a.z;
		dest.w = src.b.f;
	}
	static void go(v4 &dest, matrix_binop<binop_expand,
		basic_matrix<2, 1>, basic_matrix<2, 1> > src)
	{
		dest.x = src.a.x;
		dest.y = src.a.y;
		dest.z = src.b.x;
		dest.w = src.b.y;
	}

	// ordinary matrix multiplication
	static void go(m4 &, matrix_binop<binop_mul, m4, m4>);
	static void go(v4 &, matrix_binop<binop_mul, m4, v4>);
};

} // namespace math
} // namespace cpse
